@@ -119,6 +119,13 @@ static nsIXPCSecurityManager* gWorkerSecurityManager = nsnull;
 
 PRUintn gJSContextIndex = BAD_TLS_INDEX;
 
+static const char* sPrefsToWatch[] = {
+  "dom.max_script_run_time"
+};
+
+// The length of time the close handler is allowed to run in milliseconds.
+static PRUint32 gWorkerCloseHandlerTimeoutMS = 10000;
+
 /**
  * Simple class to automatically destroy a JSContext to make error handling
  * easier.
@@ -191,7 +198,7 @@ public:
 
     nsresult rv;
 
-    if (mWorker->mOuterHandler->HasListeners(errorStr)) {
+    if (mWorker->HasListeners(errorStr)) {
       // Construct the error event.
       nsString message;
       rv = mScriptError->GetErrorMessage(message);
@@ -212,7 +219,7 @@ public:
                                  filename, lineno);
       NS_ENSURE_SUCCESS(rv, rv);
 
-      event->SetTarget(mWorker);
+      event->SetTarget(static_cast<nsDOMWorkerMessageHandler*>(mWorker));
 
       PRBool stopPropagation = PR_FALSE;
       rv = mWorker->DispatchEvent(static_cast<nsDOMWorkerEvent*>(event),
@@ -258,9 +265,11 @@ NS_IMPL_THREADSAFE_ISUPPORTS1(nsReportErrorRunnable, nsIRunnable)
 /**
  * Used to post an expired timeout to the correct worker.
  */
-class nsDOMWorkerTimeoutRunnable : public nsRunnable
+class nsDOMWorkerTimeoutRunnable : public nsIRunnable
 {
 public:
+  NS_DECL_ISUPPORTS
+
   nsDOMWorkerTimeoutRunnable(nsDOMWorkerTimeout* aTimeout)
   : mTimeout(aTimeout) { }
 
@@ -271,6 +280,28 @@ protected:
   nsRefPtr<nsDOMWorkerTimeout> mTimeout;
 };
 
+NS_IMPL_THREADSAFE_ISUPPORTS1(nsDOMWorkerTimeoutRunnable, nsIRunnable)
+
+class nsDOMWorkerKillRunnable : public nsIRunnable
+{
+public:
+  NS_DECL_ISUPPORTS
+
+  nsDOMWorkerKillRunnable(nsDOMWorker* aWorker)
+  : mWorker(aWorker) { }
+
+  NS_IMETHOD Run() {
+    NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    mWorker->Kill();
+    return NS_OK;
+  }
+
+private:
+  nsRefPtr<nsDOMWorker> mWorker;
+};
+
+NS_IMPL_THREADSAFE_ISUPPORTS1(nsDOMWorkerKillRunnable, nsIRunnable)
+
 /**
  * This class exists to solve a particular problem: Calling Dispatch on a
  * thread pool will always create a new thread to service the runnable as long
@@ -280,29 +311,56 @@ protected:
  * currently so we cheat by using a runnable that emulates a thread. The
  * nsDOMThreadService's monitor protects the queue of events.
  */
-class nsDOMWorkerRunnable : public nsRunnable
+class nsDOMWorkerRunnable : public nsIRunnable
 {
   friend class nsDOMThreadService;
 
 public:
-  nsDOMWorkerRunnable(nsDOMWorker* aWorker)
-  : mWorker(aWorker) { }
+  NS_DECL_ISUPPORTS
 
-  virtual ~nsDOMWorkerRunnable() {
-    nsCOMPtr<nsIRunnable> runnable;
-    while ((runnable = dont_AddRef((nsIRunnable*)mRunnables.PopFront()))) {
-      // Loop until all the runnables are dead.
-    }
+  nsDOMWorkerRunnable(nsDOMWorker* aWorker)
+  : mWorker(aWorker), mCloseTimeoutInterval(0), mKillWorkerWhenDone(PR_FALSE) {
   }
 
-  void PutRunnable(nsIRunnable* aRunnable) {
-    NS_ASSERTION(aRunnable, "Null pointer!");
+  virtual ~nsDOMWorkerRunnable() {
+    ClearQueue();
+  }
 
-    NS_ADDREF(aRunnable);
+  void PutRunnable(nsIRunnable* aRunnable,
+                   PRIntervalTime aTimeoutInterval,
+                   PRBool aClearQueue) {
+    NS_ASSERTION(aRunnable, "Null pointer!");
 
     // No need to enter the monitor because we should already be in it.
 
-    mRunnables.Push(aRunnable);
+    if (NS_LIKELY(!aTimeoutInterval)) {
+      NS_ADDREF(aRunnable);
+      mRunnables.Push(aRunnable);
+    }
+    else {
+      NS_ASSERTION(!mCloseRunnable, "More than one close runnable?!");
+      if (aClearQueue) {
+        ClearQueue();
+      }
+      mCloseRunnable = aRunnable;
+      mCloseTimeoutInterval = aTimeoutInterval;
+      mKillWorkerWhenDone = PR_TRUE;
+    }
+  }
+
+  void SetCloseRunnableTimeout(PRIntervalTime aTimeoutInterval) {
+    NS_ASSERTION(aTimeoutInterval, "No timeout specified!");
+    NS_ASSERTION(aTimeoutInterval!= PR_INTERVAL_NO_TIMEOUT, "Bad timeout!");
+
+    // No need to enter the monitor because we should already be in it.
+
+    NS_ASSERTION(mWorker->GetExpirationTime() == PR_INTERVAL_NO_TIMEOUT,
+                 "Asked to set timeout on a runnable with no close handler!");
+
+    // This may actually overflow but we don't care - the worst that could
+    // happen is that the close handler could run for a slightly different
+    // amount of time and the spec leaves the time up to us anyway.
+    mWorker->SetExpirationTime(PR_IntervalNow() + aTimeoutInterval);
   }
 
   NS_IMETHOD Run() {
@@ -323,9 +381,15 @@ public:
 
     JS_SetContextPrivate(cx, mWorker);
 
+    PRBool killWorkerWhenDone;
+
     // Tell the worker which context it will be using
     if (mWorker->SetGlobalForContext(cx)) {
-      RunQueue(cx);
+      RunQueue(cx, &killWorkerWhenDone);
+
+      // Code in XPConnect assumes that the context's global object won't be
+      // replaced outside of a request.
+      JSAutoRequest ar(cx);
 
       // Remove the global object from the context so that it might be garbage
       // collected.
@@ -333,21 +397,42 @@ public:
       JS_SetContextPrivate(cx, NULL);
     }
     else {
-      // This is usually due to a parse error in the worker script...
-      JS_SetGlobalObject(cx, NULL);
-      JS_SetContextPrivate(cx, NULL);
+      {
+        // Code in XPConnect assumes that the context's global object won't be
+        // replaced outside of a request.
+        JSAutoRequest ar(cx);
+
+        // This is usually due to a parse error in the worker script...
+        JS_SetGlobalObject(cx, NULL);
+        JS_SetContextPrivate(cx, NULL);
+      }
 
       nsAutoMonitor mon(gDOMThreadService->mMonitor);
+      killWorkerWhenDone = mKillWorkerWhenDone;
       gDOMThreadService->WorkerComplete(this);
       mon.NotifyAll();
     }
 
+    if (killWorkerWhenDone) {
+      nsCOMPtr<nsIRunnable> runnable = new nsDOMWorkerKillRunnable(mWorker);
+      NS_ENSURE_TRUE(runnable, NS_ERROR_OUT_OF_MEMORY);
+
+      nsresult rv = NS_DispatchToMainThread(runnable, NS_DISPATCH_NORMAL);
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
+
     return NS_OK;
   }
 
 protected:
+  void ClearQueue() {
+    nsCOMPtr<nsIRunnable> runnable;
+    while ((runnable = dont_AddRef((nsIRunnable*)mRunnables.PopFront()))) {
+      // Loop until all the runnables are dead.
+    }
+  }
 
-  void RunQueue(JSContext* aCx) {
+  void RunQueue(JSContext* aCx, PRBool* aCloseRunnableSet) {
     PRBool operationCallbackTriggered = PR_FALSE;
 
     while (1) {
@@ -357,6 +442,19 @@ protected:
 
         runnable = dont_AddRef((nsIRunnable*)mRunnables.PopFront());
 
+        if (!runnable && mCloseRunnable) {
+          PRIntervalTime expirationTime;
+          if (mCloseTimeoutInterval == PR_INTERVAL_NO_TIMEOUT) {
+            expirationTime = mCloseTimeoutInterval;
+          }
+          else {
+            expirationTime = PR_IntervalNow() + mCloseTimeoutInterval;
+          }
+          mWorker->SetExpirationTime(expirationTime);
+
+          runnable.swap(mCloseRunnable);
+        }
+
         if (!runnable || mWorker->IsCanceled()) {
 #ifdef PR_LOGGING
           if (mWorker->IsCanceled()) {
@@ -364,6 +462,7 @@ protected:
                  static_cast<void*>(mWorker.get())));
           }
 #endif
+          *aCloseRunnableSet = mKillWorkerWhenDone;
           gDOMThreadService->WorkerComplete(this);
           mon.NotifyAll();
           return;
@@ -384,6 +483,7 @@ protected:
 
       runnable->Run();
     }
+    NS_NOTREACHED("Shouldn't ever get here!");
   }
 
   // Set at construction
@@ -391,8 +491,13 @@ protected:
 
   // Protected by mMonitor
   nsDeque mRunnables;
+  nsCOMPtr<nsIRunnable> mCloseRunnable;
+  PRIntervalTime mCloseTimeoutInterval;
+  PRBool mKillWorkerWhenDone;
 };
 
+NS_IMPL_THREADSAFE_ISUPPORTS1(nsDOMWorkerRunnable, nsIRunnable)
+
 /*******************************************************************************
  * JS environment function and callbacks
  */
@@ -402,10 +507,6 @@ DOMWorkerOperationCallback(JSContext* aCx)
 {
   nsDOMWorker* worker = (nsDOMWorker*)JS_GetContextPrivate(aCx);
 
-  // Want a strong ref here to make sure that the monitor we wait on won't go
-  // away.
-  nsRefPtr<nsDOMWorkerPool> pool;
-
   PRBool wasSuspended = PR_FALSE;
   PRBool extraThreadAllowed = PR_FALSE;
   jsrefcount suspendDepth = 0;
@@ -440,16 +541,6 @@ DOMWorkerOperationCallback(JSContext* aCx)
     }
 
     if (!wasSuspended) {
-      // Make sure we can get the monitor we need to wait on. It's possible that
-      // the worker was canceled since we checked above.
-      if (worker->IsCanceled()) {
-        NS_WARNING("Tried to suspend on a pool that has gone away");
-        JS_ClearPendingException(aCx);
-        return JS_FALSE;
-      }
-
-      pool = worker->Pool();
-
       // Make sure to suspend our request while we block like this, otherwise we
       // prevent GC for everyone.
       suspendDepth = JS_SuspendRequest(aCx);
@@ -464,9 +555,19 @@ DOMWorkerOperationCallback(JSContext* aCx)
       wasSuspended = PR_TRUE;
     }
 
-    nsAutoMonitor mon(pool->Monitor());
-    mon.Wait();
+    nsAutoMonitor mon(worker->Pool()->Monitor());
+
+    // There's a small chance that the worker was canceled after our check
+    // above in which case we shouldn't wait here. We're guaranteed not to race
+    // here because the pool reenters its monitor after canceling each worker
+    // in order to notify its condition variable.
+    if (worker->IsSuspended() && !worker->IsCanceled()) {
+      mon.Wait();
+    }
   }
+
+  NS_NOTREACHED("Should never get here!");
+  return JS_FALSE;
 }
 
 void
@@ -491,9 +592,15 @@ DOMWorkerErrorReporter(JSContext* aCx,
   }
 
   nsresult rv;
-  nsCOMPtr<nsIScriptError> scriptError =
-    do_CreateInstance(NS_SCRIPTERROR_CONTRACTID, &rv);
-  NS_ENSURE_SUCCESS(rv,);
+  nsCOMPtr<nsIScriptError> scriptError;
+
+  {
+    // CreateInstance will lock, make sure we suspend our request!
+    JSAutoSuspendRequest ar(aCx);
+
+    scriptError = do_CreateInstance(NS_SCRIPTERROR_CONTRACTID, &rv);
+    NS_ENSURE_SUCCESS(rv,);
+  }
 
   const PRUnichar* message =
     reinterpret_cast<const PRUnichar*>(aReport->ucmessage);
@@ -514,28 +621,30 @@ DOMWorkerErrorReporter(JSContext* aCx,
   if (aReport->errorNumber != JSMSG_SCRIPT_STACK_QUOTA &&
       aReport->errorNumber != JSMSG_OVER_RECURSED) {
     // Try the onerror handler for the worker's scope.
-    nsCOMPtr<nsIDOMEventListener> handler =
-      worker->mInnerHandler->GetOnXListener(NS_LITERAL_STRING("error"));
+    nsRefPtr<nsDOMWorkerScope> scope = worker->GetInnerScope();
+    NS_ASSERTION(scope, "Null scope!");
 
-    if (handler) {
+    PRBool hasListeners = scope->HasListeners(NS_LITERAL_STRING("error"));
+    if (hasListeners) {
       nsRefPtr<nsDOMWorkerErrorEvent> event(new nsDOMWorkerErrorEvent());
       if (event) {
         rv = event->InitErrorEvent(NS_LITERAL_STRING("error"), PR_FALSE, PR_TRUE,
                                    nsDependentString(message), filename,
                                    aReport->lineno);
         if (NS_SUCCEEDED(rv)) {
-          NS_ASSERTION(worker->GetInnerScope(), "Null scope!");
-          event->SetTarget(worker->GetInnerScope());
+          event->SetTarget(scope);
 
           NS_ASSERTION(worker->mErrorHandlerRecursionCount >= 0,
                        "Bad recursion count logic!");
           worker->mErrorHandlerRecursionCount++;
 
-          handler->HandleEvent(static_cast<nsDOMWorkerEvent*>(event));
+          PRBool preventDefaultCalled = PR_FALSE;
+          scope->DispatchEvent(static_cast<nsDOMWorkerEvent*>(event),
+                               &preventDefaultCalled);
 
           worker->mErrorHandlerRecursionCount--;
 
-          if (event->PreventDefaultCalled()) {
+          if (preventDefaultCalled) {
             return;
           }
         }
@@ -608,6 +717,8 @@ nsDOMThreadService::Init()
 
   obs.forget(&gObserverService);
 
+  RegisterPrefCallbacks();
+
   mThreadPool = do_CreateInstance(NS_THREADPOOL_CONTRACTID, &rv);
   NS_ENSURE_SUCCESS(rv, rv);
 
@@ -719,6 +830,8 @@ nsDOMThreadService::Cleanup()
   if (gObserverService) {
     gObserverService->RemoveObserver(this, NS_XPCOM_SHUTDOWN_OBSERVER_ID);
     NS_RELEASE(gObserverService);
+
+    UnregisterPrefCallbacks();
   }
 
   // The thread pool holds a circular reference to this service through its
@@ -749,15 +862,20 @@ nsDOMThreadService::Cleanup()
 
 nsresult
 nsDOMThreadService::Dispatch(nsDOMWorker* aWorker,
-                             nsIRunnable* aRunnable)
+                             nsIRunnable* aRunnable,
+                             PRIntervalTime aTimeoutInterval,
+                             PRBool aClearQueue)
 {
   NS_ASSERTION(aWorker, "Null pointer!");
   NS_ASSERTION(aRunnable, "Null pointer!");
 
   NS_ASSERTION(mThreadPool, "Dispatch called after 'xpcom-shutdown'!");
 
-  if (aWorker->IsCanceled()) {
-    LOG(("Will not dispatch runnable [0x%p] for canceled worker [0x%p]",
+  // Don't accept the runnable if the worker's close handler has been triggered
+  // (unless, of course, this is the close runnable as indicated by the non-0
+  // timeout value).
+  if (aWorker->IsClosing() && !aTimeoutInterval) {
+    LOG(("Will not dispatch runnable [0x%p] for closing worker [0x%p]",
          static_cast<void*>(aRunnable), static_cast<void*>(aWorker)));
     return NS_ERROR_NOT_AVAILABLE;
   }
@@ -767,14 +885,14 @@ nsDOMThreadService::Dispatch(nsDOMWorker* aWorker,
     nsAutoMonitor mon(mMonitor);
 
     if (mWorkersInProgress.Get(aWorker, getter_AddRefs(workerRunnable))) {
-      workerRunnable->PutRunnable(aRunnable);
+      workerRunnable->PutRunnable(aRunnable, aTimeoutInterval, aClearQueue);
       return NS_OK;
     }
 
     workerRunnable = new nsDOMWorkerRunnable(aWorker);
     NS_ENSURE_TRUE(workerRunnable, NS_ERROR_OUT_OF_MEMORY);
 
-    workerRunnable->PutRunnable(aRunnable);
+    workerRunnable->PutRunnable(aRunnable, aTimeoutInterval, PR_FALSE);
 
     PRBool success = mWorkersInProgress.Put(aWorker, workerRunnable);
     NS_ENSURE_TRUE(success, NS_ERROR_OUT_OF_MEMORY);
@@ -806,6 +924,23 @@ nsDOMThreadService::Dispatch(nsDOMWorker* aWorker,
   return NS_OK;
 }
 
+void
+nsDOMThreadService::SetWorkerTimeout(nsDOMWorker* aWorker,
+                                     PRIntervalTime aTimeoutInterval)
+{
+  NS_ASSERTION(aWorker, "Null pointer!");
+  NS_ASSERTION(aTimeoutInterval, "No timeout specified!");
+
+  NS_ASSERTION(mThreadPool, "Dispatch called after 'xpcom-shutdown'!");
+
+  nsAutoMonitor mon(mMonitor);
+
+  nsRefPtr<nsDOMWorkerRunnable> workerRunnable;
+  if (mWorkersInProgress.Get(aWorker, getter_AddRefs(workerRunnable))) {
+    workerRunnable->SetCloseRunnableTimeout(aTimeoutInterval);
+  }
+}
+
 void
 nsDOMThreadService::WorkerComplete(nsDOMWorkerRunnable* aRunnable)
 {
@@ -844,9 +979,11 @@ nsDOMThreadService::CreateJSContext()
     nsDOMWorkerSecurityManager::JSTranscodePrincipals,
     nsDOMWorkerSecurityManager::JSFindPrincipal
   };
-
   JS_SetContextSecurityCallbacks(cx, &securityCallbacks);
 
+  static JSDebugHooks debugHooks;
+  JS_SetContextDebugHooks(cx, &debugHooks);
+
   nsresult rv = nsContentUtils::XPConnect()->
     SetSecurityManagerForJSContext(cx, gWorkerSecurityManager, 0);
   NS_ENSURE_SUCCESS(rv, nsnull);
@@ -1267,3 +1404,53 @@ nsDOMThreadService::GetUserAgent(nsAString& aUserAgent)
                "Shouldn't call this before we have loaded strings!");
   aUserAgent.Assign(mUserAgent);
 }
+
+void
+nsDOMThreadService::RegisterPrefCallbacks()
+{
+  NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+  for (PRUint32 index = 0; index < NS_ARRAY_LENGTH(sPrefsToWatch); index++) {
+    nsContentUtils::RegisterPrefCallback(sPrefsToWatch[index], PrefCallback,
+                                         nsnull);
+    PrefCallback(sPrefsToWatch[index], nsnull);
+  }
+}
+
+void
+nsDOMThreadService::UnregisterPrefCallbacks()
+{
+  NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+  for (PRUint32 index = 0; index < NS_ARRAY_LENGTH(sPrefsToWatch); index++) {
+    nsContentUtils::UnregisterPrefCallback(sPrefsToWatch[index], PrefCallback,
+                                           nsnull);
+  }
+}
+
+// static
+int
+nsDOMThreadService::PrefCallback(const char* aPrefName,
+                                 void* aClosure)
+{
+  NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+  if(!strcmp(aPrefName, "dom.max_script_run_time")) {
+    // We assume atomic 32bit reads/writes. If this assumption doesn't hold on
+    // some wacky platform then the worst that could happen is that the close
+    // handler will run for a slightly different amount of time.
+    PRUint32 timeoutMS =
+      nsContentUtils::GetIntPref(aPrefName, gWorkerCloseHandlerTimeoutMS);
+
+    // We must have a timeout value, 0 is not ok. If the pref is set to 0 then
+    // fall back to our default.
+    if (timeoutMS) {
+      gWorkerCloseHandlerTimeoutMS = timeoutMS;
+    }
+  }
+  return 0;
+}
+
+// static
+PRUint32
+nsDOMThreadService::GetWorkerCloseHandlerTimeoutMS()
+{
+  return gWorkerCloseHandlerTimeoutMS;
+}
